/**************************************************************************
*  Copyright (c) 2017-2018 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Based on an example from NXP. Therefore partial copyright:
*  Copyright(C) 2012 NXP Semiconductors
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  17.03.2017  mifi  First Version, tested with an Embedded Artists 
*                    LPC4088 Quickstart Board and QSB4088com.
**************************************************************************/
#if defined(USE_BOARD_BF4088)
#define __FATFS_DISKIO_C__

/*=======================================================================*/
/*  Includes                                                             */
/*=======================================================================*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tal.h"
#include "diskio.h"

#include "fsmci_cfg.h"
#include "ffconf.h"
#include "chip.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all global Data                                        */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static os_sema_t FSSema;

static DSTATUS DiskStatus = STA_NOINIT;

static CARD_HANDLE_T *hCard;

SDMMC_CARD_T sdCardInfo;

static volatile int32_t sdcWaitExit = 0;
static SDMMC_EVENT_T *event;
static volatile Status  eventResult = ERROR;

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/* Delay callback for timed SDIF/SDMMC functions */
STATIC void waitMs(uint32_t time)
{
   OSTimeWait(time);
}

/**
 * @brief   Sets up the event driven wakeup
 * @param   pEvent : Event information
 * @return  Nothing
 */
STATIC void setupEvWakeup(void *pEvent)
{
#ifdef SDC_DMA_ENABLE
   /* Wait for IRQ - for an RTOS, you would pend on an event here with a IRQ based wakeup. */
   NVIC_ClearPendingIRQ(DMA_IRQn);
#endif
   event = (SDMMC_EVENT_T *)pEvent;
   sdcWaitExit = 0;
   eventResult = ERROR;
#ifdef SDC_DMA_ENABLE
   NVIC_EnableIRQ(DMA_IRQn);
#endif /*SDC_DMA_ENABLE*/
}

/**
 * @brief   A better wait callback for SDMMC driven by the IRQ flag
 * @return  0 on success, or failure condition (Nonzero)
 */
STATIC uint32_t waitEvIRQDriven(void)
{
   /* Wait for event, would be nice to have a timeout, but keep it  simple */
   while (sdcWaitExit == 0) {}
   if (eventResult) {
      return 0;
   }

   return 1;
}

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  ff_memalloc                                                          */
/*                                                                       */
/*  Allocate a memory block                                              */
/*                                                                       */
/*  In    : msize                                                        */
/*  Out   : none                                                         */
/*  Return: Returns pointer to the allocated memory block                */
/*************************************************************************/
void *ff_memalloc (UINT msize)
{
   return(malloc(msize));
} /* ff_memalloc */

/*************************************************************************/
/*  ff_memfree                                                           */
/*                                                                       */
/*  Free a memory block                                                  */
/*                                                                       */
/*  In    : mblock                                                       */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void ff_memfree (void *mblock)
{
   free(mblock);
} /* ff_memfree */

/*************************************************************************/
/*  ff_cre_syncobj                                                       */
/*                                                                       */
/*  Create a Synchronization Object                                      */
/*                                                                       */
/*  In    : vol,  corresponding logical drive being processed.           */
/*          sobj, pointer to return the created sync object.             */
/*  Out   : none                                                         */
/*  Return: 0 = Error / 1 = OK                                           */
/*************************************************************************/
int ff_cre_syncobj (BYTE vol, FF_SYNC_t *sobj)
{
   (void)vol;
   
   /* Init the semaphore */
   OSSemaCreate(&FSSema, 1, 1);
   *sobj = &FSSema;

   return(1);
} /* ff_cre_syncobj */

/*************************************************************************/
/*  ff_del_syncobj                                                       */
/*                                                                       */
/*  Delete a Synchronization Object                                      */
/*                                                                       */
/*  In    : sobj, sync object tied to the logical drive to be deleted.   */
/*  Out   : none                                                         */
/*  Return: 0 = Error / 1 = OK                                           */
/*************************************************************************/
int ff_del_syncobj (FF_SYNC_t sobj)
{
   /* Reset the semaphore */
   OSSemaDelete(sobj);
   
   return(1);
} /* ff_del_syncobj */

/*************************************************************************/
/*  ff_req_grant                                                         */
/*                                                                       */
/*  Request Grant to Access the Volume                                   */
/*                                                                       */
/*  In    : sobj, sync object to wait.                                   */
/*  Out   : none                                                         */
/*  Return: 1 = Got a grant / 0 =  Could not get a grant                 */
/*************************************************************************/
int ff_req_grant (FF_SYNC_t sobj)
{
   OSSemaWait(sobj, OS_WAIT_INFINITE);
   
   return(1);
} /* ff_req_grant */

/*************************************************************************/
/*  ff_rel_grant                                                         */
/*                                                                       */
/*  Request Grant to Access the Volume                                   */
/*                                                                       */
/*  In    : sobj, Sync object to be signaled.                            */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void ff_rel_grant (FF_SYNC_t sobj)
{
   OSSemaSignal(sobj);
} /* ff_rel_grant */

/*************************************************************************/
/*  disk_initialize                                                      */
/*                                                                       */
/*  Initialize a Drive                                                   */
/*                                                                       */
/*  In    : pdrv, physical drive nmuber to identify the drive            */
/*  Out   : none                                                         */
/*  Return: Status of Disk Functions                                     */
/*************************************************************************/
DSTATUS disk_initialize (uint8_t pdrv)
{
   static uint8_t InitDone = 0;
   
   if (0 == pdrv)
   {
      if (0 == InitDone)
      {
         InitDone = 1;

         /* Use P0.7 for card detect */   
         Chip_IOCON_PinMuxSet(LPC_IOCON, 0, 7, (IOCON_FUNC0 | IOCON_MODE_PULLUP | IOCON_DIGMODE_EN));
         Chip_GPIO_SetPinDIRInput(LPC_GPIO, 0, 7);
         
         memset(&sdCardInfo, 0, sizeof(sdCardInfo));
         sdCardInfo.evsetup_cb   = setupEvWakeup;
         sdCardInfo.waitfunc_cb  = waitEvIRQDriven;
         sdCardInfo.msdelay_func = waitMs;

         Chip_IOCON_PinMuxSet(LPC_IOCON, 1, 2,  (IOCON_FUNC2 | IOCON_MODE_INACT));  /*lint !e845*/
         Chip_IOCON_PinMuxSet(LPC_IOCON, 1, 3,  (IOCON_FUNC2 | IOCON_MODE_INACT));  /*lint !e845*/
         Chip_IOCON_PinMuxSet(LPC_IOCON, 1, 5,  (IOCON_FUNC2 | IOCON_MODE_INACT | IOCON_DIGMODE_EN));  /*lint !e845*/
         Chip_IOCON_PinMuxSet(LPC_IOCON, 1, 6,  (IOCON_FUNC2 | IOCON_MODE_INACT | IOCON_DIGMODE_EN));  /*lint !e845*/
         Chip_IOCON_PinMuxSet(LPC_IOCON, 1, 7,  (IOCON_FUNC2 | IOCON_MODE_INACT | IOCON_DIGMODE_EN));  /*lint !e845*/
         Chip_IOCON_PinMuxSet(LPC_IOCON, 1, 11, (IOCON_FUNC2 | IOCON_MODE_INACT));  /*lint !e845*/
         Chip_IOCON_PinMuxSet(LPC_IOCON, 1, 12, (IOCON_FUNC2 | IOCON_MODE_INACT));  /*lint !e845*/

         Chip_SDC_Init(LPC_SDC);
      }
      
      /* Initialize the Card Data Strucutre */
      hCard = FSMCI_CardInit();
      OSTimeDly(100);

      /* Reset */
      DiskStatus = STA_NOINIT;

      FSMCI_CardInsertWait(hCard); /* Wait for card to be inserted */

      /* Enumerate the card once detected. Note this function may block for a little while. */
      if (!FSMCI_CardAcquire(hCard)) {
         DEBUGOUT("Card Acquire failed...\r\n");
         return(DiskStatus);
      }

      DiskStatus &= ~STA_NOINIT;
   }
   else
   {
      DiskStatus = STA_NOINIT;
   }      

   return(DiskStatus);
} /* disk_initialize */

/*************************************************************************/
/*  disk_status                                                          */
/*                                                                       */
/*  Get Drive Status                                                     */
/*                                                                       */
/*  In    : pdrv, physical drive nmuber to identify the drive            */
/*  Out   : none                                                         */
/*  Return: Status of Disk Functions                                     */
/*************************************************************************/
DSTATUS disk_status (uint8_t pdrv)
{
   if (0 == pdrv)
   {
      /* Check if card is available */
      if (0 == Chip_GPIO_ReadPortBit(LPC_GPIO, 0, 7))
      {
         DiskStatus &= ~STA_NODISK;
      }
      else
      {
         DiskStatus |= STA_NODISK;
      }
   }
   else
   {
      DiskStatus = STA_NOINIT;
   } 

   return(DiskStatus);
} /* disk_status */

/*************************************************************************/
/*  disk_removed                                                         */
/*                                                                       */
/*  Medium was removed                                                   */
/*                                                                       */
/*  In    : pdrv, physical drive nmuber to identify the drive            */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void disk_removed (BYTE pdrv)
{
   if (0 == pdrv)
   {
      DiskStatus |= (STA_NODISK | STA_NOINIT);
      
      /* Power-off */
      Chip_SDC_PowerControl(LPC_SDC, SDC_POWER_OFF, 0);      
   }
} /* disk_removed */

/*************************************************************************/
/*  disk_read                                                            */
/*                                                                       */
/*  Read Sector(s)                                                       */
/*                                                                       */
/*  In    : pdrv,   physical drive nmuber to identify the drive          */
/*          buff,   data buffer to store read data                       */
/*          sector, sector address in LBA                                */
/*          count,  number of sectors to read                            */
/*  Out   : none                                                         */
/*  Return: Results of Disk Functions                                    */
/*************************************************************************/
DRESULT disk_read (BYTE pdrv, BYTE* buff, DWORD sector, UINT count)
{
   if (pdrv || !count) {
      return RES_PARERR;
   }
   if (DiskStatus & STA_NOINIT) {
      return RES_NOTRDY;
   }

   if (FSMCI_CardReadSectors(hCard, buff, (int32_t)sector, (int32_t)count)) {
      return RES_OK;
   }

   return RES_ERROR;
} /* disk_read */

/*************************************************************************/
/*  disk_write                                                           */
/*                                                                       */
/*  Write Sector(s)                                                      */
/*                                                                       */
/*  In    : pdrv,   physical drive nmuber to identify the drive          */
/*          buff,   data to be written                                   */
/*          sector, sector address in LBA                                */
/*          count,  number of sectors to write                           */
/*  Out   : none                                                         */
/*  Return: Results of Disk Functions                                    */
/*************************************************************************/
DRESULT disk_write (BYTE pdrv, const BYTE* buff, DWORD sector, UINT count)
{
   if (pdrv || !count) {
      return RES_PARERR;
   }
   if (DiskStatus & STA_NOINIT) {
      return RES_NOTRDY;
   }

   if ( FSMCI_CardWriteSectors(hCard, (void *) buff, (int32_t)sector, (int32_t)count)) {
      return RES_OK;
   }

   return RES_ERROR;
} /* disk_write */

/*************************************************************************/
/*  disk_ioctl                                                           */
/*                                                                       */
/*  In    : pdrv, physical drive nmuber to identify the drive            */
/*          cmd,  control code                                           */
/*          buff, buffer to send/receive control data                    */
/*  Out   : none                                                         */
/*  Return: Results of Disk Functions                                    */
/*************************************************************************/
DRESULT disk_ioctl (uint8_t pdrv, uint8_t cmd, void *buff)
{
   DRESULT res;
   BYTE *ptr = buff;

   if (pdrv) {
      return RES_PARERR;
   }
   if (DiskStatus & STA_NOINIT) {
      return RES_NOTRDY;
   }

   res = RES_ERROR;

   switch (cmd) {
   case CTRL_SYNC:   /* Make sure that no pending write process */
      if (FSMCI_CardReadyWait(hCard, 50)) {
         res = RES_OK;
      }
      break;

   case GET_SECTOR_COUNT:  /* Get number of sectors on the disk (DWORD) */
      *(DWORD *) buff = FSMCI_CardGetSectorCnt(hCard);
      res = RES_OK;
      break;

   case GET_SECTOR_SIZE:   /* Get R/W sector size (WORD) */
      *(WORD *) buff = (WORD)FSMCI_CardGetSectorSz(hCard);
      res = RES_OK;
      break;

   case GET_BLOCK_SIZE:/* Get erase block size in unit of sector (DWORD) */
      *(DWORD *) buff = FSMCI_CardGetBlockSz(hCard);
      res = RES_OK;
      break;

   case MMC_GET_TYPE:      /* Get card type flags (1 byte) */
      *ptr = (BYTE)FSMCI_CardGetType(hCard);
      res = RES_OK;
      break;

   case MMC_GET_CSD:    /* Receive CSD as a data block (16 bytes) */
      *((uint32_t *) buff + 0) = FSMCI_CardGetCSD(hCard, 0);
      *((uint32_t *) buff + 1) = FSMCI_CardGetCSD(hCard, 1);
      *((uint32_t *) buff + 2) = FSMCI_CardGetCSD(hCard, 2);
      *((uint32_t *) buff + 3) = FSMCI_CardGetCSD(hCard, 3);
      res = RES_OK;
      break;

   case MMC_GET_CID:    /* Receive CID as a data block (16 bytes) */
      *((uint32_t *) buff + 0) = FSMCI_CardGetCID(hCard, 0);
      *((uint32_t *) buff + 1) = FSMCI_CardGetCID(hCard, 1);
      *((uint32_t *) buff + 2) = FSMCI_CardGetCID(hCard, 2);
      *((uint32_t *) buff + 3) = FSMCI_CardGetCID(hCard, 3);
      res = RES_OK;
      break;

   case MMC_GET_SDSTAT:/* Receive SD status as a data block (64 bytes) */
      if (FSMCI_CardGetState(hCard, (uint8_t*)buff)) {   /*lint !e826*/
         res = RES_OK;
      }
      break;

   default:
      res = RES_PARERR;
      break;
   }

   return res;
} /* disk_ioctl */

/*************************************************************************/
/*  get_fattime                                                          */
/*                                                                       */
/*  Gets Time from RTC                                                   */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Time                                                         */
/*************************************************************************/
uint32_t get_fattime (void)
{
   return(0);
}

#ifdef SDC_DMA_ENABLE
/**
 * @brief   GPDMA interrupt handler sub-routine
 * @return  Nothing
 */
void DMA_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   
   eventResult = Chip_GPDMA_Interrupt(LPC_GPDMA, event->DmaChannel);
   sdcWaitExit = 1;
   NVIC_DisableIRQ(DMA_IRQn);
   
   TAL_CPU_IRQ_EXIT();
}
#endif /*SDC_DMA_ENABLE*/

/**
 * @brief   SDC interrupt handler sub-routine
 * @return  Nothing
 */
void SDIO_IRQHandler(void)
{
   int32_t Ret;

   TAL_CPU_IRQ_ENTER();
   
#ifdef SDC_DMA_ENABLE
   Ret = Chip_SDMMC_IRQHandler(LPC_SDC, NULL,0,NULL,0);
#else
   if(event->Index < event->Size) {
      if(event->Dir) { /* receive */
         Ret = Chip_SDMMC_IRQHandler(LPC_SDC, NULL,0,(uint8_t*)event->Buffer,&event->Index);
      }
      else {
         Ret = Chip_SDMMC_IRQHandler(LPC_SDC, (uint8_t*)event->Buffer,&event->Index,NULL,0);
      }
   }
   else {
      Ret = Chip_SDMMC_IRQHandler(LPC_SDC, NULL,0,NULL,0);
   }
#endif
   if(Ret < 0) {
      eventResult = ERROR;
      sdcWaitExit = 1;
   }
#ifndef SDC_DMA_ENABLE
   else if(Ret == 0) {
      eventResult = SUCCESS;
      sdcWaitExit = 1;
   }
#endif

   TAL_CPU_IRQ_EXIT();
}

uint8_t SD_Detect (void)
{
   uint8_t status = SD_PRESENT;
   
   
   if (disk_status(0) & STA_NODISK)
   {
      status = SD_NOT_PRESENT;
   }
   
   return(status);
}

#endif /* USE_BOARD_BF4088 */

/*** EOF ***/
